﻿using System;
using Android.App;
using Android.Widget;
using System.Threading;
using System.Threading.Tasks;
using System.Collections.Generic;
using System.Net.Http;
using HttpsBug;

namespace HttpClientBug {
    
    [Activity(Label = "Http Client Bug", MainLauncher = true, Icon = "@drawable/icon")]
    public class MainActivity : Activity {

        HttpClient commonClient;

        protected override void OnCreate(Android.OS.Bundle bundle) {
            base.OnCreate(bundle);

            // Set our view from the "main" layout resource
            SetContentView(Resource.Layout.Main);

            // NOTE 1: Uncomment this line to stop HttpClient request hang

//            System.Net.ServicePointManager.DefaultConnectionLimit = 100;

            var button = FindViewById<Button>(Resource.Id.myButton);
            button.Text = "Start";
            button.Click += async delegate {

                Console.WriteLine("---------Running Test");

                button.Text = "Running...";
                button.Enabled = false;

                commonClient = new HttpClient() { Timeout = TimeSpan.FromSeconds(30) };

                // Start download tasks that should be cancelled
                var ctsCancel = new CancellationTokenSource();
                var downloadTaskCancel = MultiDownload(ctsCancel.Token, "Cancel");

                // Start download tasks that should not be cancelled
                var ctsDoNotCancel = new CancellationTokenSource();
                var downloadTaskDoNotCancel = MultiDownload(ctsDoNotCancel.Token, "Do Not Cancel");               

                // NOTE 2: Remove the cancel operation to stop the hang

                // Cancel first set of tasks
                ctsCancel.Cancel();

                // Wait for cancelled tasks to complete
                await downloadTaskCancel;

                // Wait for non-cancelled tasks to complete
                await downloadTaskDoNotCancel;

                button.Enabled = true;
                button.Text = "Start";

                Console.WriteLine("---------Test Complete");
            };
        }

        async Task MultiDownload(CancellationToken ct, string tag) {
            var tasks = new List<Task>();
            for (int i = 0; i < 10; i++) {
                tasks.Add(Download(ct, i, tag));
            }
            await Task.WhenAll(tasks);
        }

        async Task Download(CancellationToken ct, int i, string tag) {
            Console.WriteLine($"Starting '{tag}' {i}");
            try {
                
                // NOTE 3: Swap localClient for commonClient in the request to stop the hang

                using (var localClient = new HttpClient() { Timeout = TimeSpan.FromSeconds(30) })
                using (var response = await commonClient.GetAsync("https://download.mozilla.org/?product=firefox-30.0-SSL&os=linux&lang=ach", HttpCompletionOption.ResponseHeadersRead, ct)) {
                    if (response.IsSuccessStatusCode) {
                        Console.WriteLine($"Finished '{tag}' {i}");
                    } else {
                        Console.WriteLine($"Failed '{tag}' {i}");
                    }
                }
            } catch (TaskCanceledException) {
                Console.WriteLine($"Task Cancelled '{tag}' {i}");
            } catch (Exception ex) {                
                Console.WriteLine($"Exception '{tag}' {i} - {ex.Message}");
                Console.WriteLine(ex.StackTrace);
            }
        }

    }
}

